import CParser;
import java.util.Vector;
import java.lang.Integer;

//
//
// CBoolFkt
//
//
public class CBoolFkt 
{
	// Ausdruck
	public String sExpression;
	// Anzahl benutzter Variablen
	public int nVars;
	// Anzahl Eintraege in der Wertetabelle
	public int nAmount;
	// benutzte Variablen als String
	public String sActiveVars;
	// ueberpruefter Input
	public String sValidInput;

	// Klassenzugehoerigkeiten
	public boolean isK0, isK1, isKsd, isKlinear, isKmonoton;
	// Normalformen
	public StringBuffer sVDNF, sVKNF, sANF, sMDNF;

	// Wertetabelle, Index als Bitfeld der Parameter
	public boolean[] arResults;
	// ANF als Tabelle entsprechend der Bit-Indizes
	private boolean[] arANF;


	// Berechnungsobjekt instanziieren
	public CParser Parser = new CParser();


	// Ausdruck formatieren und auf gueltige Syntax ueberpruefen
	public int Format()
	{
		// an den Parser weiterreichen
		int Result = Parser.Format(sExpression);

		// Anzahl Variablen und Groesse Wertetabelle ermitteln
		nVars = Parser.nMinVars;
		nAmount = 1<<nVars;

		// benutzte Variablen und validierten Input speichern
		sActiveVars = Parser.sActiveVars.toString();
		sValidInput = Parser.sValidInput.toString();

		// Ergebnis der Formatierung durch den Parser uebermitteln
		return Result;
	}


	// Wertetabelle erstellen
	public void DoCalc()
	{
		// Tabelle allokieren
		arResults = new boolean[nAmount];

		// alle Eintraege berechnen
		for (int i=0; i<nAmount; i++)
		{
			// Variablen belegen
			for (int run=0; run<nVars; run++)
				Parser.Vars[Parser.sActiveVars.charAt(run)] = IsBitSet(i,nVars-run-1);

			// Ausdruck mittels der Belegung berechnen
			arResults[i]=Parser.Evaluate();
		}

		// Normalformen bestimmen
		DoNF();
		// Klassenzugehoerigkeiten ermitteln
		DoClass();
	}


	// Normalformen bestimmen
	private void DoNF()
	{
		// erstellen
		sVDNF = new StringBuffer();
		sVKNF = new StringBuffer();
		sMDNF = new StringBuffer();
		sANF = new StringBuffer();
		
	
		// komplette Wertetabelle durchlaufen
		for (int i=0; i<nAmount; i++)
			// DNF
			if (arResults[i])
			{
				// die einzelne Variablen auf Zugehoerigkeit testen
				for (int run=0; run<nVars; run++)
				{
					// ggf. negieren
					if (!IsBitSet(i,nVars-run-1))
						sVDNF.append("!");
					// Variable hinzufuegen
					sVDNF.append(Parser.sActiveVars.charAt(run));
				}
				sVDNF.append(" | ");
			}
			// KNF
			else
			{
				sVKNF.append("(");
				// die einzelne Variablen auf Zugehoerigkeit testen
				for (int run=0; run<nVars; run++)
				{
					// ggf. negieren
					if (IsBitSet(i,nVars-run-1))
						sVKNF.append("!");
					// Variable hinzufuegen
					sVKNF.append(Parser.sActiveVars.charAt(run)+"|");
				}
				// letztes "|" entfernen
				sVKNF.setLength(sVKNF.length()-1);

				sVKNF.append(") & ");
			}

		// letztes " | " entfernen
		if (sVDNF.length()>0)
			sVDNF.setLength(sVDNF.length()-3);
		else
		{
			// leere DNF, also Kontradiktion 
			// (mit 3 Dummy-Space fuer folgende Entfernung bei KNF)
			sVDNF = new StringBuffer("0 (Kontradiktion)   ");
			sVKNF = new StringBuffer("0 (Kontradiktion)   ");
			sMDNF = new StringBuffer("0 (Kontradiktion)   ");
			sANF  = new StringBuffer("0 (Kontradiktion)   ");
		}

		// letztes " & " entfernen
		if (sVKNF.length()>0)
			sVKNF.setLength(sVKNF.length()-3);
		else
		{
			// leere KNF, also Tautologie
			sVKNF = new StringBuffer("1 (Tautologie)");
			sVDNF = new StringBuffer("1 (Tautologie)");
			sMDNF = new StringBuffer("1 (Tautologie)");
			sANF  = new StringBuffer("1 (Tautologie)");
		}

		// Minimierung der DNF durchfuehren, wenn DNF keine Sonderform ist (Tautologie etc.)
		if (sMDNF.length()==0)
			DoQuineMcClusky();
		else
			// die ANF ist dann zur DNF identisch und somit bereits ermittelt worden
			return;


		// ANF ermitteln

		// Tabelle allokieren
		arANF = new boolean[nAmount];

		// f(0,..,0) bestimmt, ob ANF negiert wird
		arANF[0]=arResults[0];
		if (arANF[0])
			sANF.append("1 + ");

		// alle Variablenanzahlen durchlaufen
		for (int nVarCount=1; nVarCount<=nVars; nVarCount++)
			// jede Belegung mit dieser Anzahl Variablen bearbeiten
			for (int i=nAmount-1; i>0; i--)
				if (BitCount(i)==nVarCount)
				{
					// mit jeder Konjunktion, die Bestandteil der gerade
					// zu bearbeitenden Konjunktion, wird antivalent
					// verknuepft
					arANF[i] = arResults[i];
					for (int run=0; run<i; run++)
						if ((i|run)==i)
							arANF[i] ^= arResults[run];

					// der Zeichenkette hinzufuegen
					if (arANF[i])
					{
						for (int run=0; run<nVars; run++)
							if (IsBitSet(i,nVars-run-1))
								sANF.append(Parser.sActiveVars.charAt(run));
						sANF.append(" + ");
					}
				}

		// ggf. letztes " + " entfernen
		sANF.setLength(sANF.length()-3);
	}


	// Minimierung der DNF nach Quine-McClusky
	private void DoQuineMcClusky()
	{
		// dynamische Arrays:
		//   die Konjunktionen selber
		Vector arConjunctions = new Vector();
		//   Bitfelder, die angeben, welche der Bestandteile der Konjunktionen auch gueltig sind
		Vector arUsedVars = new Vector();
		//   Markierung, die anzeigt, dass diese Konjunktion ersetzt wird
		Vector arMarked = new Vector();

		// zuerst werden alle Variablen in den Konjunktionen gebraucht
		int nUsedVarsMask = nAmount-1;

		// die vollstaendige DNF erfassen
		for (int i=0; i<nAmount; i++)
			if (arResults[i])
			{
				arConjunctions.addElement(new Integer(i));
				arUsedVars.addElement(new Integer(nUsedVarsMask));
				arMarked.addElement(new Boolean(false));
			}

		// keine weitere Optimierung mehr moeglich
		if (arConjunctions.size()<=1)
		{
			// die VDNF ist auch die MDNF
			sMDNF = new StringBuffer(sVDNF.toString());
			// isch haben fertisch
			return;
		}


		// alle Raenge durcharbeiten
		for (int level=0; level<nVars; level++)
		{
			// wenn nur noch eine Konjunktion uebrigbleibt, ist die mDNF gefunden
			int nConjunctions = arConjunctions.size();
			if (nConjunctions<=1)
				break;

			// jede Konjunktion mit allen ihr in der Wertetabellen folgenden vergleichen
			for (int checknr=0; checknr<nConjunctions-1; checknr++)
				for (int comparenr=checknr+1; comparenr<nConjunctions; comparenr++)
					// nur die interessant, die auf den gleichen Variablen aufbauen
					if (((Integer)arUsedVars.elementAt(  checknr)).intValue()==
					    ((Integer)arUsedVars.elementAt(comparenr)).intValue())
					{
						// Exponenten holen (die zeigen an, ob Variablen negiert sind oder nicht)
						int checkval   = ((Integer)arConjunctions.elementAt(  checknr)).intValue();
						int compareval = ((Integer)arConjunctions.elementAt(comparenr)).intValue();
						// Unterschiede feststellen
						int DifferMask = XorBits(checkval,compareval);

						// nur eine Variable darf sich zwischen beiden unterscheiden
						if (BitCount(DifferMask)==1)
						{
							// die benutzten Konjunktionen streichen
							arMarked.setElementAt(new Boolean(true),  checknr);
							arMarked.setElementAt(new Boolean(true),comparenr);

							// neue Konjunktion erstellen und hinzufuegen
							//    Exponenten ermitteln
							int NewConj = AndBits(checkval,compareval);
							//    Variablenbenutzung markieren
							int NewMask = XorBits(DifferMask,((Integer)arUsedVars.elementAt(checknr)).intValue());

							//    und hinzufuegen
							arConjunctions.addElement(new Integer(NewConj));
							arUsedVars.addElement(new Integer(NewMask));
							arMarked.addElement(new Boolean(false));
						}
					}

			// Vector-Groesse aktualisieren
			nConjunctions = arConjunctions.size();

			// Ueberdeckungen (Redundanzen) entfernen
			// dazu alle nicht markierten Konjunktionen miteinander vergleichen
			for (int checknr=0; checknr<nConjunctions-1; checknr++)
				if (((Boolean)arMarked.elementAt(checknr)).booleanValue()==false)
					for (int comparenr=checknr+1; comparenr<nConjunctions; comparenr++)
						if (((Boolean)arMarked.elementAt(comparenr)).booleanValue()==false)
						{
							// die Exponenten der Variablen holen
							int checkval   = ((Integer)arConjunctions.elementAt(  checknr)).intValue();
							int compareval = ((Integer)arConjunctions.elementAt(comparenr)).intValue();
			
							// die benutzten Variablen als Maske abspeichern
							int checkmask   = ((Integer)arUsedVars.elementAt(  checknr)).intValue();
							int comparemask = ((Integer)arUsedVars.elementAt(comparenr)).intValue();

							// gemeinsame Variablen ermitteln
							int maskand = AndBits(checkmask,comparemask);

							// falls identische Variablen benutzt
							if (checkmask == comparemask)
							{
								// und die auch noch die gleichen Exponenten haben
								if (checkval == compareval)
									// dann sind die Konjunktionen identisch, eine kann also entfallen
									arMarked.setElementAt(new Boolean(true), comparenr);
							}
							else
							{
								// nachschauen, ob eine Maske alle gemeinsamen Variablen enthaelt
								// und dann pruefen, ob diese auch die gleichen Exponenten haben
								// wenn ja, dann ueberdeckt eine Konjunktion die andere vollstaendig

								// ueberdeckt "check" die Konjunktion "compare" ?
								if ((maskand == checkmask)&&(AndBits(compareval,maskand) == checkval))
									arMarked.setElementAt(new Boolean(true), comparenr);

								// ueberdeckt "compare" die Konjunktion "check" ?
								if ((maskand == comparemask)&&(AndBits(checkval,maskand) == compareval))
									arMarked.setElementAt(new Boolean(true), checknr);
							}
						}


			// alle markierten Konjunktionen entfernen, dabei von hinten anfangen,
			// um nicht bei der Vector-Groesse in Probleme zu kommen
			for (int cleanup=nConjunctions-1; cleanup>=0; cleanup--)
				if (((Boolean)arMarked.elementAt(cleanup)).booleanValue()==true)
				{
					arConjunctions.removeElementAt(cleanup);
					arUsedVars.removeElementAt(cleanup);
					arMarked.removeElementAt(cleanup);
				}
		}


		// Ueberdeckungstabelle erstellen, somit aus der verkuerzten DNF die minimierte erstellen
		// Achtung: es koennen mehrere mDNF zu einer vDNF existieren !

		// die Bedeutung von arMarked wird jetzt umgekehrt: TRUE bedeutet ab sofort, dass
		// diese Konjunktion unbedingt notwendig ist und nicht weggelassen werden darf
		// (nach dem letzten CleanUp sind alle nun verbliebenen mit FALSE markiert)
/*
		boolean bAllDone = false;

		boolean[] arNotMinimized = new boolean[nAmount];
		for (int i=0; i<nAmount; i++)
			arNotMinimized[i] = arResults[i];
				
		while (!bAllDone)
		{
			int[] arCoveredCount = new int[nAmount];
			int[] arCoveredBy = new int[nAmount];

			bAllDone = true;

			for (int i=0; i<nAmount; i++)
			{
				arCoveredCount[i] = 0;
				arCoveredBy[i] = -1;

				if (arNotMinimized[i])
					for (int nConj=0; nConj<arConjunctions.size(); nConj++)
						if (((Boolean)arMarked.elementAt(nConj)).booleanValue()==false)
						{
							int checkval  = ((Integer)arConjunctions.elementAt(nConj)).intValue();
							int checkmask = ((Integer)arUsedVars.elementAt(nConj)).intValue();

							if (AndBits(checkmask,i) == checkval)
							{
								arCoveredCount[i]++;
								arCoveredBy[i] = nConj;

								bAllDone = false;
							}
						}
			}

			if (bAllDone)
				break;
		

			boolean bFoundNew = false;

			for (int i=0; i<nAmount; i++)
				if (arCoveredCount[i]==1)
				{
					arMarked.setElementAt(new Boolean(true), arCoveredBy[i]);
					bFoundNew = true;
				}


			for (int i=0; i<nAmount; i++)
				if (arNotMinimized[i])
					for (int nConj=0; nConj<arConjunctions.size(); nConj++)
						if (((Boolean)arMarked.elementAt(nConj)).booleanValue()==true)
						{
							int checkval  = ((Integer)arConjunctions.elementAt(nConj)).intValue();
							int checkmask = ((Integer)arUsedVars.elementAt(nConj)).intValue();

							if (AndBits(checkmask,i) == checkval)
								arNotMinimized[i]=false;
						}


			int nLowestVarCount = 100;
			int nLowestVarPos = -1;

			if (!bFoundNew)
			{
				for (int nConj=0; nConj<arConjunctions.size(); nConj++)
					if (((Boolean)arMarked.elementAt(nConj)).booleanValue()==false)
						if (BitCount(((Integer)arUsedVars.elementAt(nConj)).intValue())<nLowestVarCount)
						{
							nLowestVarCount = BitCount(((Integer)arUsedVars.elementAt(nConj)).intValue());
							nLowestVarPos = nConj;
						}

				if (nLowestVarPos!=-1)
					arMarked.setElementAt(new Boolean(true), nLowestVarPos);
			}
		}
*/
		

		// Ausgabe der MDNF
		for (int i=0; i<arConjunctions.size(); i++)
//			if (((Boolean)arMarked.elementAt(i)).booleanValue()==true)
			{
				// benutzte Variablen
				int nUsed = ((Integer)arUsedVars.elementAt(i)).intValue();
				// die dazugehoerigen Exponenten
				int nConj = ((Integer)arConjunctions.elementAt(i)).intValue();

				// alle moeglichen Variablen durchlaufen
				for (int run=0; run<nVars; run++)
					// nur wenn die auch in der Konjunktion benutzt wird
					if (IsBitSet(nUsed,nVars-run-1))
					{
						// gemaess Exponent ggf. negieren
						if (!IsBitSet(nConj,nVars-run-1))
							sMDNF.append("!");

						// den Variablennamen ergaenzen
						sMDNF.append(Parser.sActiveVars.charAt(run));
					}

				// und die naechste Konjunktion vorbereiten
				sMDNF.append(" | ");
			}

		// letztes " | " entfernen
		sMDNF.setLength(sMDNF.length()-3);
	}


	// Klassenzugehoerigkeiten ermitteln
	private void DoClass()
	{
		// K0: f(0,..,0)=0
		isK0=!arResults[0];
		// K1: f(1,..,1)=1
		isK1=arResults[nAmount-1];

		// f(a,..,z)=f(!a,..,!z)
		isKsd = true;
		// nach einem Gegenbeispiel suchen
		// nur erste Haelfte durchlaufen, dabei mit der 2. vergleichen
		for (int i=0; i<nAmount/2; i++)
			if (arResults[i]==arResults[nAmount-i-1])
			{
				isKsd = false;
				break;
			}

		// ANF darf nur Terme enthalten, die aus genau einer Variablen bestehen
		isKlinear = true;
		// f(0,..,0) ueberspringen, da ohne Einfluss
		for (int i=1; i<nAmount; i++)
			if (arANF[i])
				if (BitCount(i)>1)
				{
					isKlinear=false;
					break;
				}

		// fuer a0<=b0, a1<=b1, ... ist f(a0,a1,..)<=f(b0,b1,..)
		isKmonoton = true;
		// wieder ein Gegenbeispiel suchen
		// wenn f(b)=0, darf kein f(a)=1 sein (a<b)
		for (int i=1; i<nAmount; i++)
			// nur f(b)=0 untersuchen
			if (!arResults[i])
				for (int j=0; j<i; j++)
					// alle a<b ermitteln, f(a) darf aber nicht 1 sein
					if (((i|j)==i)&&arResults[j])
					{
						isKmonoton = false;
						break;
					}
	}


	// testen, ob ein spezielles Bit gesetzt ist (von rechts mit 0 beginnend)
	public boolean IsBitSet(int Number, int Bit)
	{
		return (Number&(1<<Bit))>0;
	}	

	// logische bitweise AND-Verknuepfung zweier Integer-Werte
	public int AndBits(int a, int b)
	{
		// Ergebnis
		int Result = 0;
		// aktuell bearbeitetes Bit
		int CurrBit = 0;

		// a und b vollstaendig bearbeiten
		// dabei immer Bit 0 betrachten, danach die restliche Zahl um 1 Bit nach rechts schieben
		while ((a>0)||(b>0))
		{
			// nur wenn beide Bit 0 von a und b gesetzt sind, wird im Ergebnis das
			// Bit an der Stelle CurrBit (von rechts gesehen) gesetzt
			if (IsBitSet(a,0)&&IsBitSet(b,0))
				Result+=1<<CurrBit;

			// naechstes Bit im Ergebnis
			CurrBit++;

			// a und b um 1 Bit nach rechts schieben
			a>>=1;
			b>>=1;
		}

		return Result;
	}

	// logische bitweise XOR-Verknuepfung zweier Integer-Werte
	public int XorBits(int a, int b)
	{
		// Ergebnis
		int Result = 0;
		// aktuell bearbeitetes Bit
		int CurrBit = 0;
		
		// a und b vollstaendig bearbeiten
		// dabei immer Bit 0 betrachten, danach die restliche Zahl um 1 Bit nach rechts schieben
		while ((a>0)||(b>0))
		{
			// nur wenn beide Bit 0 von a und b verschieden sind, wird im Ergebnis das
			// Bit an der Stelle CurrBit (von rechts gesehen) gesetzt
			if (IsBitSet(a,0)!=IsBitSet(b,0))
				Result+=1<<CurrBit;

			// naechstes Bit im Ergebnis
			CurrBit++;

			// a und b um 1 Bit nach rechts schieben
			a>>=1;
			b>>=1;
		}

		return Result;
	}

	// Anzahl gesetzter Bits ermitteln
	public int BitCount(int Number)
	{
		int Result = 0;

		// bitweise nach rechts schieben, niederwertigstes Bit pruefen
		while (Number>0)
		{
			// ist niederwertigstes Bit gesetzt ?
			if ((Number&1)!=0)
				Result++;

			// um 1 Bit nach rechts schieben
			Number>>=1;
		}

		return Result;
	}
}