//
//
// CParser
//
//
public class CParser
{
	// Konstanten zur uebersichtlichen Fehlerbehandlung
	public final int OK=0, NOTOKEN=1, UNKNOWNTOKEN=2,
					 NOVARS=3, VAREXPECTED=4,
		             BRACKETEXPECTED=5, BRACKETNOTALLOWED=6;
	// max. Anzahl Variablen
	public final int nMaxVars = 12;
	// Anzahl benutzter Variablen
	public int nMinVars;
	// genaue Namen der benutzten Variablen, alphabetisch sortiert
	public StringBuffer sActiveVars;
	// als gueltig erkannter Teil der Aussage
	public StringBuffer sValidInput;
	// Wertebelegung der Variablen per ASCII-Namen, also z.B. Vars['a']=true
	public boolean[] Vars = new boolean[128];

	// zur besseren Uebersicht die Variablennamen
	final char _A='a', _B='b', _C='c', _D='d',
			   _E='e', _F='f', _G='g', _H='h',
			   _I='i', _J='j', _K='k', _L='l',
	// moegliche Operationen
			   _OR='|', _AND='&', _XOR='+', _EQU='=', _NOT='!',
			   _LOG='>', _LB='(', _RB=')', _TRUE='1', _FALSE='0',
	// Markierung des Aussagenendes
			   _END=0;

	// Woerter, die durch Symbole ersetzt werden
	final String[] arLongNames = { "or","oder","and","und",
								   "xor","not","nicht","=>" };
	// Symbole, die Woerter ersetzen
	final char[] arLongToken = { '|','|','&','&','+','!','!','>' };

	// die Aussage zur weiteren Bearbeitung
	private StringBuffer sExpression;
	// Zeiger bei der Zerlegung der Aussage
	private int iExpressionCount;
	// aktuell erkanntes Zeichen
	private char cToken;


	// passt Aussage der Sprache des Parsers an, erkennt ungueltige Syntax
	public int Format(String Expression)
	{
		// Variable erwarten
		boolean bForceVar = true;
		// Schachtelungstiefe durch Klammerung mitzaehlen
		int nBracketCount = 0;

		// neue Zeichenkette aus Expression aufbauen
		sValidInput = new StringBuffer();

		// es sollte schon was drinstehen ...
		if (Expression.length()==0)
			return NOTOKEN;

		// in Kleinbuchstaben umwandeln
		Expression = Expression.toLowerCase();

		// alle Woerter (z.B. AND, OR etc. durch ihre Symbole ersetzen
		StringBuffer sNoLongNames = new StringBuffer(Expression);

		// alle moeglichen Woerter abtesten
		for (int i=0; i<arLongNames.length; i++)
			// wirklich jedes Auftreten ersetzen
			while(sNoLongNames.toString().indexOf(arLongNames[i])!=-1)
			{
				// Zwischenspeicher zum Aufbau der neuen Aussage
				StringBuffer sTemp = new StringBuffer();

				// genaue Adresse des Worter speichern
				int index = sNoLongNames.toString().indexOf(arLongNames[i]);

				// nur wenn nicht als erstes (als 1. nur bei NOT relevant)
				if (index>0)
					// Aussage vor dem Wort unveraendert uebernehmen
					sTemp.append(sNoLongNames.toString().substring(0,index));

				// Wort ersetzen
				sTemp.append(arLongToken[i]);

				// nur wenn nicht als letztes (somit falsche Syntax !)
				if ((index+arLongNames[i].length())<sNoLongNames.length())
					// Aussage nach dem Wort unveraendert uebernehmen
					sTemp.append(sNoLongNames.toString().substring(index+arLongNames[i].length()));

				// neue Aussage uebernehmen und weiterbearbeiten
				sNoLongNames = sTemp;
			}

		// alle Woerter sind ersetzt
		Expression = sNoLongNames.toString();

		// alle Zeichen durchlaufen
		for (int i=0; i<Expression.length(); i++)
			// Token ueberpruefen
			switch(Expression.charAt(i))
			{
			// Leerzeichen ignorieren
			case ' ': continue;

			// zulaessige Token
			// zuerst Variablen / Konstanten
			case _FALSE:; case _TRUE:;
			case _A:; case _B:;	case _C:; case _D:;
			case _E:; case _F:;	case _G:; case _H:;
			case _I:; case _J:;	case _K:; case _L:
				// es darf jedes Token folgen
				bForceVar = false;
				// ok bis hierher
				sValidInput.append(Expression.charAt(i));
				break;

			// unaerer Operator der Negation
			case _NOT:;
				// Variable muss folgen
				bForceVar = true;
				// ok bis hierher
				sValidInput.append(Expression.charAt(i));
				break;

			// binaere Operatoren
			case _AND:; case _OR:; case _XOR:; case _LOG:; case _EQU:;
				// Variable erwartet ? wenn ja, dann Fehler
				if (bForceVar)
					return VAREXPECTED;
				// Variable muss folgen
				bForceVar = true;
				// ok bis hierher
				sValidInput.append(Expression.charAt(i));
				break;

			// oeffnende linke Klammer
			case _LB:;
				// Verschachtelung mitzaehlen
				nBracketCount++;
				// Variable muss folgen
				bForceVar = true;
				// ok bis hierher
				sValidInput.append(Expression.charAt(i));
				break;

			// schliessende rechte Klammer
			case _RB:;
				// muss erlaubt sein, es darf weder Variable erwartet werden,
				// noch duerfen mehr rechte als linke Klammern auftauchen
				if (bForceVar||(nBracketCount==0))
					return BRACKETNOTALLOWED;
				// "zurueckschachteln" (Copyright auf dieses Wort beantragt)
				nBracketCount--;

				// ok bis hierher
				sValidInput.append(Expression.charAt(i));
				break;
			// ungueltiges Token
			default: return UNKNOWNTOKEN;
			}

		// alle Klammern ordentlich aufgeloest ?
		if (nBracketCount!=0)
			return BRACKETEXPECTED;

		// es darf keine Variable mehr erwartet werden
		if (bForceVar)
			return VAREXPECTED;

		// Anzahl Variablen zaehlen
		nMinVars=0;
		// vorkommende Variablen mit ihrem Namen festhalten
		sActiveVars = new StringBuffer();
		// moegliche Namen begrenzen
		String sAllVars = "abcdefghijkl";
		// auf alle moeglichen Variablen ueberpruefen
		int i;
		for (i=0; i<nMaxVars; i++)
			// taucht sie in der Aussage auf ?
	        if (sValidInput.toString().indexOf(sAllVars.charAt(i))!=-1)
			{
				// ja, also vermerken
				nMinVars++;
				sActiveVars.append(sAllVars.charAt(i));
			}

		// konstante Ausdruecke sind nicht erlaubt
		if (nMinVars==0)
			return NOVARS;

		// die formatierte Aussage an Aufrufer zurueckgeben
		Expression = sValidInput.toString();
		// Formatierung abgeschlossen, fuer die anschliessende Auswertung vormerken
		sExpression = sValidInput;
		// Ende fuer die Auswertung markieren
		sExpression.append(_END);

		// fehlerfrei bearbeitet
		return OK;
	}


	// Start der Auswertung
	public boolean Evaluate()
	{
		// Aussage festhalten
		sExpression = sValidInput;

		// Zeiger auf Start
		iExpressionCount = 0;

		// ... und auswerten
		return DoEqu(true);
	}


	// Aequivalenz berechnen
	// Parameter: muss cToken aktualisiert werden ?
	private boolean DoEqu(boolean get)
	{
		// ggf. neues cToken holen
		boolean left = DoLog(get);

		// Aequivalenz-Bloecke berechnen
		for (;;)
			if (cToken==_EQU)
				left=(DoLog(true)==left);
			else
				return left;
	}

	// Implikation berechnen
	// Parameter: muss cToken aktualisiert werden ?
	private boolean DoLog(boolean get)
	{
		// ggf. neues cToken holen
		boolean left = DoOr(get);

		// Implikations-Bloecke berechnen
		for (;;)
			if (cToken==_LOG)
				// a>b=!a|b
				left=DoOr(true)||!left;
			else
				return left;
	}

	// Disjunktion berechnen
	// Parameter: muss cToken aktualisiert werden ?
	private boolean DoOr(boolean get)
	{
		// ggf. neues cToken holen
		boolean left = DoXor(get);

		// Disjunktions-Bloecke berechnen
		for (;;)
			if (cToken==_OR)
				left|=DoXor(true);
			else
				return left;
	}

	// Antivalenz berechnen
	// Parameter: muss cToken aktualisiert werden ?
	private boolean DoXor(boolean get)
	{
		// ggf. neues cToken holen
		boolean left = DoAnd(get);

		// Antivalenz-Bloecke berechnen
		for (;;)
			if (cToken==_XOR)
				left^=DoAnd(true);
			else
				return left;
	}

	// Konjunktion berechnen
	// Parameter: muss cToken aktualisiert werden ?
	private boolean DoAnd(boolean get)
	{
		// ggf. neues cToken holen
		boolean left = DoPrimary(get);

		// Konjunktions-Bloecke berechnen
		for (;;)
		{
			switch (cToken)
			{
			// ja, explizit markiert, rekursiv absteigen und Aussage berechnen
			case _AND:
				left&=DoPrimary(true); break;

			// ja, impliziert durch Kurzschreibweise (a&b=ab)
			case _A:; case _B:;	case _C:; case _D:;
			case _E:; case _F:;	case _G:; case _H:;
			case _I:; case _J:;	case _K:; case _L:;
			case _LB:; case _NOT:
			// rekursiv absteigen und Aussage berechnen
			left&=DoPrimary(false); break;
			default:
			// fertich glaubich
				return left;
			}
		}
	}


	// atomares Zeichen ermitteln
	// Parameter: muss cToken aktualisiert werden ?
	private boolean DoPrimary(boolean get)
	{
		// ggf. neues Token auslesen
		if (get)
			GetToken();

		// Ergebnis-Zwischenspeicher
		boolean temp;

		// Token auswerten
		switch (cToken)
		{
		// Variableninhalt zurueckgeben
		case _A:; case _B:;	case _C:; case _D:;
		case _E:; case _F:;	case _G:; case _H:;
		case _I:; case _J:;	case _K:; case _L:;
			temp = Vars[cToken];
			// naechtes Token auslesen
			GetToken();
			return temp;

		// unaere Negation
		case _NOT:
			// inverse Aussage des naechsten Tokens zurueckgeben
			return !DoPrimary(true);

		// Konstante "TRUE"=1
		case _TRUE:
			// naechtes Token auslesen
			GetToken();
			// "TRUE" zurueckgeben
			return true;

		// Konstante "FALSE"=0
		case _FALSE:
			// naechtes Token auslesen
			GetToken();
			// "FALSE" zurueckgeben
			return false;

		// oeffnende Klammer
		case _LB:
			// die Klammer auswerten
			temp = DoEqu(true);
			// naechtes Token auslesen
			GetToken();
			// Klammerinhalt zurueckgeben
			return temp;

		// Dummy-Zeile, damit der Compiler nicht meckert
		default: return true;
		}
	}

	// einzelnes Zeichen aus der Aussage auslesen
	private void GetToken()
	{
		// Zeichen lesen, Index weiterbewegen
		cToken = sExpression.charAt(iExpressionCount++);
	}


};